//
//  GraphView.m
//  LuaConsole
//
//  Created by 永田 央 on 16/01/09.
//  Copyright 2016 __MyCompanyName__. All rights reserved.
//

#import "GraphView.h"
#include "screen.h"

#import <Carbon/Carbon.h>

static NSLock *sLock;       //  The lock to access shared memory
static NSRect sInvalRect;   //  The rectangle to be drawn in the main thread

static NSTimer *sTimer;     //  Interval timer to activate the main thread

/*  Frame buffer (Cocoa-specific) */
static u_int32_t *sFrameBuffer;

/*  Color palette  */
#if BYTES_PER_PIXEL == 2
static pixel_t sPalette[65536];
#endif

@implementation GraphView

#define CREATE_FONT 0

#if CREATE_FONT
- (void)createFontData
{
	//  The following code snippet was used in generating the bitmap data
	static uint32_t *sBitmapBuffer;
	NSDictionary *attr;
	NSGraphicsContext *sTextContext;
	NSBitmapImageRep *rep;
	unichar ch[10];
	int i, j, k, count, xx, yy;
	u_int16_t b;
	u_int32_t uc, eu, idx;
	FILE *fp1, *fp2, *fp3;
	NSString *s;
	char buf[64], *p;
	char buf2[16][80];
	
	/*  See graph.h for details  */
	static u_int16_t sConvTable[65536];
	static u_int8_t sFontData[16*188];
	static u_int8_t sKanjiData[32*11844];
	
	sBitmapBuffer = (uint32_t *)calloc(sizeof(uint32_t), 1024);
	rep = [[NSBitmapImageRep alloc] initWithBitmapDataPlanes:(uint8_t **)&sBitmapBuffer pixelsWide:32 pixelsHigh:32 bitsPerSample:8 samplesPerPixel:4 hasAlpha:YES isPlanar:NO colorSpaceName:NSCalibratedRGBColorSpace bytesPerRow:0 bitsPerPixel:0];
	sTextContext = [NSGraphicsContext graphicsContextWithBitmapImageRep:rep];
	[sTextContext setShouldAntialias:NO];
	[NSGraphicsContext saveGraphicsState];
	[NSGraphicsContext setCurrentContext:sTextContext];
	attr = [NSDictionary dictionaryWithObjectsAndKeys:[NSFont fontWithName:@"Sazanami-Gothic-Regular" size:15], NSFontAttributeName, [NSColor whiteColor], NSForegroundColorAttributeName, nil];
	//	attr = [NSDictionary dictionaryWithObjectsAndKeys:[NSFont userFixedPitchFontOfSize:14], NSFontAttributeName, [NSColor whiteColor], NSForegroundColorAttributeName, nil];
	
	fp2 = fopen("fontdata.ascii", "w");
	fp3 = fopen("../../euc-jis-2004-with-char-u8.txt", "r");
	count = 0;
	for (i = 0; i < 65536; i++)
		sConvTable[i] = 0xffff;
	while (fgets(buf, sizeof buf, fp3) != NULL) {
		j = k = 0;
		for (i = 0; buf[i] != 0; i++) {
			if (buf[i] == '\t') {
				if (j == 0)
					j = i + 3;
				else if (k == 0) {
					k = i + 3;
					break;
				}
			}
		}
		eu = strtol(buf + j, NULL, 16);
		uc = strtol(buf + k, &p, 16);
		if (*p == '+')
			uc = uc * 65536 + strtol(p + 1, &p, 16);  /*  Two-character pair  */
		if (uc >= 65536) {
			count++;
			printf("%d: Cannot convert: %s", count, buf);
			continue;
		}
		if (eu >= 0x8fa1a1) {
			if (eu >= 0x8feea1) {
				idx = 10434 + (((eu >> 8) & 0xff) - 0xee) * 94 + ((eu & 0xff) - 0xa1);
			} else {
				idx = 9024 + (((eu >> 8) & 0xff) - 0xa1) * 94 + ((eu & 0xff) - 0xa1);
			}
		} else if (eu >= 0xa1a1) {
			idx = 188 + (((eu >> 8) & 0xff) - 0xa1) * 94 + ((eu & 0xff) - 0xa1);
		} else if (eu >= 0x8ea1) {
			idx = 94 + ((eu & 0xff) - 0xa1);
		} else if (eu >= 0x21 && eu <= 0x7e) {
			idx = eu - 0x21;
		} else continue; /* Silently skip */
		sConvTable[uc] = idx;
	}
	fclose(fp3);
	memset(sFontData, 0, sizeof(sFontData));
	memset(sKanjiData, 0, sizeof(sKanjiData));
	for (uc = 0; uc < 65536; uc++) {
		idx = sConvTable[uc];
		if (idx == 0xffff)
			continue;
		ch[0] = uc;
		s = [[NSString alloc] initWithCharacters:ch length:1];
		memset(sBitmapBuffer, 0, sizeof(sBitmapBuffer[0]) * 1024);
		[s drawWithRect:NSMakeRect(0, 0, 16, 32) options:NSStringDrawingUsesLineFragmentOrigin attributes:attr];
		[s release];
		for (yy = 0; yy < 16; yy++) {
			b = 0;
			if (idx < 188) {
				for (xx = 0; xx < 8; xx++) {
					if (sBitmapBuffer[(yy + 6) * 32 + xx] != 0)
						b |= (1 << xx);
				}
				sFontData[idx * 16 + yy] = b;
			} else {
				for (xx = 0; xx < 16; xx++) {
					if (sBitmapBuffer[(yy + 6) * 32 + xx] != 0)
						b |= (1 << xx);
				}
				j = (idx - 188) * 32 + yy * 2;
				sKanjiData[j] = b & 0xff;
				sKanjiData[j + 1] = (b >> 8) & 0xff;
			}
		}
	}
	for (idx = 0; idx < 12032; idx++) {
		if (idx < 188) {
			i = idx % 8;
			if (i == 0) {
				for (j = 0; j < 8; j++) {
					k = j + idx;
					if (k < 94)
						eu = k + 0x21;
					else
						eu = k + 0x8ea1;
					fprintf(fp2, "%04X%s", eu, (j == 7 ? "\n" : "     "));
				}
				memset(buf2, ' ', sizeof(buf2));
			}
			for (yy = 0; yy < 16; yy++) {
				b = sFontData[idx * 16 + yy];
				for (xx = 0; xx < 8; xx++) {
					buf2[yy][i * 9 + xx] = ((b & (1 << xx)) != 0 ? '*' : ' ');
				}
			}
			if (i == 7 || idx == 187) {
				for (yy = 0; yy < 16; yy++) {
					buf2[yy][70] = '\n';
					buf2[yy][71] = 0;
					fputs(buf2[yy], fp2);
				}
			}
		} else {
			i = (idx - 188) % 4;
			if (i == 0) {
				for (j = 0; j < 4; j++) {
					k = j + idx;
					if (k < 9024) {
						eu = 0xa1a1 + ((k - 188) / 94) * 256 + (k - 188) % 94;
						fprintf(fp2, "%04X%s", eu, (j == 3 ? "\n" : "             "));
					} else {
						if (idx < 10434)
							eu = 0x8fa1a1 + ((k - 9024) / 94) * 256 + (k - 9024) % 94;
						else
							eu = 0x8feea1 + ((k - 10434) / 94) * 256 + (k - 10434) % 94;
						fprintf(fp2, "%06X%s", eu, (j == 3 ? "\n" : "           "));
					}
				}
				memset(buf2, ' ', sizeof(buf2));
			}
			for (yy = 0; yy < 16; yy++) {
				b = sKanjiData[(idx - 188) * 32 + yy * 2] + (sKanjiData[(idx - 188) * 32 + yy * 2 + 1] * 256);
				for (xx = 0; xx < 16; xx++) {
					buf2[yy][i * 17 + xx] = ((b & (1 << xx)) != 0 ? '*' : ' ');
				}
			}
			if (i == 3 || idx == 12031) {
				for (yy = 0; yy < 16; yy++) {
					buf2[yy][70] = '\n';
					buf2[yy][71] = 0;
					fputs(buf2[yy], fp2);
				}
			}
		}
	}
	[NSGraphicsContext restoreGraphicsState];
	fclose(fp2);
	
	/*  Write the font data to file  */
	fp2 = fopen("fontdata.c", "w");
	fprintf(fp2, "unsigned short gConvTable[%d] = {\n  ", (int)sizeof(sConvTable)/2);
	for (uc = 0; uc < 65536; uc++) {
		fprintf(fp2, "%d", sConvTable[uc]);
		if (uc == 65535)
			fputs("\n};\n", fp2);
		else if (uc % 16 == 15)
			fputs(",\n  ", fp2);
		else fputc(',', fp2);		
	}
	fprintf(fp2, "unsigned char gFontData[%d] = {\n  ", (int)sizeof(sFontData));
	for (i = 0; i < sizeof(sFontData); i++) {
		fprintf(fp2, "%d", sFontData[i]);
		if (i == sizeof(sFontData) - 1)
			fputs("\n};\n", fp2);
		else if (i % 16 == 15)
			fputs(",\n  ", fp2);
		else fputc(',', fp2);
	}
	fprintf(fp2, "unsigned char gKanjiData[%d] = {\n  ", (int)sizeof(sKanjiData));
	for (i = 0; i < sizeof(sKanjiData); i++) {
		fprintf(fp2, "%d", sKanjiData[i]);
		if (i == sizeof(sKanjiData) - 1)
			fputs("\n};\n", fp2);
		else if (i % 32 == 31)
			fputs(",\n  ", fp2);
		else fputc(',', fp2);
	}
	fclose(fp2);
	fp1 = fopen("fontdata.bin", "wb");
	
	i = NSSwapHostIntToLittle(sizeof(sConvTable) / 2);
	fwrite(&i, 1, 4, fp1);
	i = NSSwapHostIntToLittle(sizeof(sFontData));
	fwrite(&i, 1, 4, fp1);
	i = NSSwapHostIntToLittle(sizeof(sKanjiData));
	fwrite(&i, 1, 4, fp1);
	
	for (uc = 0; uc < 65536; uc++) {
		/*  Convert to little endian  */
		u_int16_t si;
		((char *)(&si))[0] = sConvTable[uc] & 0xff;
		((char *)(&si))[1] = (sConvTable[uc] >> 8) & 0xff;
	}
	fwrite(sConvTable, 1, sizeof(sConvTable), fp1);
	fwrite(sFontData, 1, sizeof(sFontData), fp1);
	fwrite(sKanjiData, 1, sizeof(sKanjiData), fp1);
	fclose(fp1);
	
}
#endif

- (void)awakeFromNib
{
	NSRect frame = [self frame];
	
	my_width = frame.size.width - 8;
	my_height = frame.size.height - 8;
	
	[self setBoundsOrigin:NSMakePoint(-4, -4)];
	
#if CREATE_FONT
	[self createFontData];
#endif
	
	sLock = [[NSLock alloc] init];
	sTimer = [NSTimer scheduledTimerWithTimeInterval:0.05 target:self selector:@selector(timerHandler:) userInfo:nil repeats:YES];
}

- (void)drawRect:(NSRect)dirtyRect
{
	[[NSColor blackColor] set];
	NSRectFill(dirtyRect);
	if (bitmapRep != nil) {
		[bitmapRep drawAtPoint:NSZeroPoint];
	}
}

- (void)timerHandler:(NSTimer *)aTimer
{
	int x1, x2, y1, y2, x, y, ofs, ofs2;
	int cx1, cx2, cy1, cy2, mx1, mx2, my1, my2;
	pixel_t tpix;
	
	/*  If the BASIC thread is not initialized yet, then do nothing  */
	if (sFrameBuffer == NULL)
		return;
	
	if (sInvalRect.size.width > 0 && sInvalRect.size.height > 0) {
		NSRect rect;
		int width, width2;
		[sLock lock];
		rect = sInvalRect;
		x1 = rect.origin.x;
		y1 = rect.origin.y;
		x2 = rect.origin.x + rect.size.width;
		y2 = rect.origin.y + rect.size.height;
		if (my_graphic_mode == 1) {
			mx1 = my_origin_x;
			mx2 = mx1 + my_width / 2;
			my1 = my_origin_y;
			my2 = my1 + my_height / 2;
		} else {
			mx1 = my1 = 0;
			mx2 = my_width;
			my2 = my_height;
		}
		if (x1 < mx1)
			x1 = mx1;
		if (x2 > mx2)
			x2 = mx2;
		if (y1 < my1)
			y1 = my1;
		if (y2 > my2)
			y2 = my2;
		if (my_show_cursor) {
			cx1 = my_cursor_x * 8;
			cx2 = cx1 + 8;
			cy1 = my_cursor_y * 16;
			cy2 = cy1 + 16;
		} else cx1 = cx2 = cy1 = cy2 = -1;
		if (my_graphic_mode == 1) {
			ofs = y1 * my_width + x1;
			ofs2 = (y1 * 2 + my_origin_y) * my_width + (x1 * 2 + my_origin_x);
			width = x2 - x1;
			width2 = (x2 - x1) * 2;
		} else {
			ofs = ofs2 = y1 * my_width + x1;
			width = width2 = x2 - x1;
		}
		if (my_graphic_mode == 1) {
			/*  (2x2) expansion  */
			for (y = y1; y < y2; y++) {
				for (x = x1; x < x2; x++) {
					u_int32_t pal;
					tpix = my_textscreen[ofs];
					if (tpix == 0 && x < cx2 && x >= cx1 && y < cy2 && y >= cy1)
						tpix = RGBFLOAT(0.5, 0.5, 0.5);  /*  cursor color: 50% gray  */
					if (tpix == 0)
						tpix = my_graphscreen[ofs];
#if BYTES_PER_PIXEL == 2
					pal = sPalette[tpix];
#else
					pal = tpix;
#endif
					sFrameBuffer[ofs2] = pal;
					sFrameBuffer[ofs2 + 1] = pal;
					sFrameBuffer[ofs2 + my_width] = pal;
					sFrameBuffer[ofs2 + my_width + 1] = pal;
					ofs++;
					ofs2 += 2;
				}
				ofs += my_width - width;
				ofs2 += my_width * 2 - width2;
			}			
		} else {
			for (y = y1; y < y2; y++) {
				for (x = x1; x < x2; x++) {
					tpix = my_textscreen[ofs];
					if (tpix == 0 && x < cx2 && x >= cx1 && y < cy2 && y >= cy1)
						tpix = RGBFLOAT(0.5, 0.5, 0.5);
					if (tpix == 0)
						tpix = my_graphscreen[ofs];
#if BYTES_PER_PIXEL == 2
					sFrameBuffer[ofs] = sPalette[tpix];
#else
					sFrameBuffer[ofs] = tpix;
#endif
					ofs++;
				}
				ofs += my_width - width;
			}
		}
		sInvalRect = NSMakeRect(0, 0, 0, 0);
		[sLock unlock];
		//  Create a new NSBitmapImageRep
		//  (It is necessary to create a new instance every time when the bitmap is
		//  externally modified.)
		if (bitmapRep != nil)
			[bitmapRep release];
		bitmapRep = [[NSBitmapImageRep alloc] initWithBitmapDataPlanes:(unsigned char **)&sFrameBuffer pixelsWide:my_width pixelsHigh:my_height bitsPerSample:8 samplesPerPixel:4 hasAlpha:YES isPlanar:NO colorSpaceName:NSCalibratedRGBColorSpace bytesPerRow:0 bitsPerPixel:32];
		if (my_graphic_mode == 1) {
			rect.origin.x *= 2;
			rect.origin.y *= 2;
			rect.size.width *= 2;
			rect.size.height *= 2;
			rect = NSMakeRect(0, 0, my_width, my_height);
		}
		rect.origin.y = my_height - (rect.origin.y + rect.size.height);
		[self setNeedsDisplayInRect:rect];
	}
}

//  Key inputs
#define kKeyBufferMax 128
static u_int16_t sKeyBuffer[kKeyBufferMax];
static int sKeyBufferBase, sKeyBufferCount;

- (BOOL)acceptsFirstResponder
{
	return YES;
}

static void
s_RegisterKeyDown(u_int16_t *ch)
{
	while (*ch != 0) {
		if (sKeyBufferCount < kKeyBufferMax) {
			sKeyBuffer[(sKeyBufferBase + sKeyBufferCount) % kKeyBufferMax] = *ch;
			sKeyBufferCount++;
		}
		ch++;
	}
}

/*  TODO: is it possible to interface with anthy for Japanese input?  */
/*  (https://osdn.jp/projects/anthy/)  */

- (void)keyDown: (NSEvent *)theEvent
{
	unichar ch[6];
	ch[0] = [[theEvent characters] characterAtIndex:0];
	ch[1] = 0;
	switch (ch[0]) {
		case NSUpArrowFunctionKey: ch[0] = 0x1b; ch[1] = '['; ch[2] = 'A'; ch[3] = 0; break;
		case NSDownArrowFunctionKey: ch[0] = 0x1b; ch[1] = '['; ch[2] = 'B'; ch[3] = 0; break;
		case NSLeftArrowFunctionKey: ch[0] = 0x1b; ch[1] = '['; ch[2] = 'D'; ch[3] = 0; break;
		case NSRightArrowFunctionKey: ch[0] = 0x1b; ch[1] = '['; ch[2] = 'C'; ch[3] = 0; break;
		case NSDeleteFunctionKey: ch[0] = 127; break;
		case NSBreakFunctionKey: ch[0] = 27; break;
		default: break;
	}
	s_RegisterKeyDown(ch);
}

@end

#pragma mark ====== Limonene Interface ======

void
lm_init_framebuffer(void)
{
	int i, j, k;
	
	/*  Wait for my_width and my_height to be initilized in awakeFromNib  */
	while (my_width == 0) {
		usleep(50000);
	}
	
	sFrameBuffer = (u_int32_t *)calloc(sizeof(u_int32_t), my_width * my_height);
	
	sInvalRect = NSMakeRect(0, 0, 0, 0);
	
#if BYTES_PER_PIXEL == 2
	for (i = 0; i < 32; i++) {
		for (j = 0; j < 64; j++) {
			for (k = 0; k < 32; k++) {
				sPalette[i * 2048 + j * 32 + k] = RGBFLOAT(i / 31.0, j / 63.0, k / 31.0);
			}
		}
	}
#endif
}

void
lm_lock(void)
{
	[sLock lock];
}

void
lm_unlock(void)
{
	[sLock unlock];
}

/*  Request redraw of a part of the screen  */
void
lm_redraw(int16_t x, int16_t y, int16_t width, int16_t height)
{
	NSRect aRect = NSMakeRect(x, y, width, height);
	lm_lock();
	sInvalRect = NSUnionRect(sInvalRect, aRect);
	lm_unlock();
}

/*  Get the key input.   */
int
lm_getch_raw(void)
{
	int ch;
	if (sKeyBufferCount > 0) {
		[sLock lock];
		ch = sKeyBuffer[sKeyBufferBase];
		sKeyBufferBase = (sKeyBufferBase + 1) % kKeyBufferMax;
		sKeyBufferCount--;
		[sLock unlock];		
	} else ch = -1;
	return ch;
}

/*  Dummy entry  */
int
tty_puts(const char *s)
{
	return 0;
}
