/*
 *  raspi_framebuffer.c
 *  LuaConsole
 *
 *  Created by Toshi Nagata on 16/02/07.
 *  Copyright 2016 Toshi Nagata. All rights reserved.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>

#ifndef __CONSOLE__
#if __BAREMETAL__
#include <fb.h>
#else
#include <linux/fb.h>
#include <sys/mman.h>
#endif
#endif

#include "screen.h"

#pragma mark ====== Lua Interface ======

static int s_original_bpp;

static char *s_framebuffer;
static int s_fbfd = 0;

static void
s_pb_raspi_framebuffer(void)
{
#ifndef __CONSOLE__
#if __BAREMETAL__
	fb_init(0, 0);
	my_width = fb_width;
	my_height = fb_height;
	s_framebuffer = (char *)fb_addr;
#else
	struct fb_var_screeninfo vinfo;
	struct fb_fix_screeninfo finfo;
	
	/*  Open frame buffer device  */
	s_fbfd = open("/dev/fb0", O_RDWR);
	if (!s_fbfd) {
		printf("Error: cannot open framebuffer device.\n");
		return;
	}
	if (ioctl(s_fbfd, FBIOGET_VSCREENINFO, &vinfo)) {
		printf("Error reading variable information.\n");
		return;
	}
	printf("Frame buffer %dx%d, %dbpp\n", vinfo.xres, vinfo.yres, vinfo.bits_per_pixel );
	
	s_original_bpp = vinfo.bits_per_pixel;
	vinfo.bits_per_pixel = 16;
	if (ioctl(s_fbfd, FBIOPUT_VSCREENINFO, &vinfo)) {
		printf("Error setting variable information.\n");
		return;
	}
	
	my_width = vinfo.xres;
	my_height = vinfo.yres;

	if (ioctl(s_fbfd, FBIOGET_FSCREENINFO, &finfo)) {
		printf("Error reading fixed information.\n");
		return;
	}

	/*  Map the memory buffer  */
	s_framebuffer = (char *)mmap(0, finfo.smem_len, PROT_READ | PROT_WRITE, MAP_SHARED, s_fbfd, 0);
	if (s_framebuffer == (char *)-1) {
		printf("Failed to mmap.\n");
		return;
	}
	printf("mmap'ed buffer %p\n", mmap);
#endif
#endif
}

void
lm_init_framebuffer(void)
{
#ifndef __CONSOLE__
	s_pb_raspi_framebuffer();
#endif
}

void
lm_lock(void)
{
	/*  Nothing to do  */
}

void
lm_unlock(void)
{
	/*  Nothing to do  */
}

/*  Request redraw of a part of the screen  */
void
lm_redraw(int16_t x, int16_t y, int16_t width, int16_t height)
{
	int xx, yy, ofs, cx1, cx2, cy1, cy2;
	if (x < 0)
		x = 0;
	if (x > my_width)
		x = my_width;
	if (y < 0)
		y = 0;
	if (y > my_height)
		y = my_height;
	if (width < 0)
		width = 0;
	if (height < 0)
		height = 0;
	if (x + width > my_width)
		width = my_width - x;
	if (y + height > my_height)
		height = my_height - y;
	ofs = y * my_width + x;
	if (my_show_cursor) {
		cx1 = my_cursor_x * 8 - x;
		cx2 = cx1 + 8;
		cy1 = my_cursor_y * 16 - y;
		cy2 = cy1 + 16;
	} else {
		cx1 = cx2 = cy1 = cy2 = 0;
	}
#if __BAREMETAL__
	s_framebuffer = (char *)fb_addr;  /*  May change because of double buffering  */
#endif
	for (yy = 0; yy < height; yy++) {
		for (xx = 0; xx < width; xx++) {
			u_int16_t c = my_textscreen[ofs];
			if (c == 0 && xx >= cx1 && xx < cx2 && yy >= cy1 && yy < cy2)
				c = 0x7be7;  /*  Cursor color (50% gray)  */
			if (c == 0)
				c = my_graphscreen[ofs];
			((u_int16_t *)s_framebuffer)[ofs] = c;
			ofs++;
		}
		ofs += my_width - width;
	}
}

/*  Key input.  */
/*  The standard input should be set to the RAW mode  */
int
lm_getch_raw(void)
{
	return tty_getch();
}
