/*
 *  graph.c
 *  LuaConsole
 *
 *  Created by 永田 央 on 16/01/09.
 *  Copyright 2016 __MyCompanyName__. All rights reserved.
 *
 */

#include "screen.h"
#include "graph.h"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <math.h>
#include <alloca.h>

static u_int8_t s_strokeflag = 1, s_fillflag = 0;
static pixel_t s_gcolor = RGBFLOAT(1, 1, 1), s_fillcolor = 0;

#define LUA_AUTOINT(L, n) (lua_isinteger(L, n) ? lua_tointeger(L, n) : (int)floor(luaL_checknumber(L, n) + 0.5))

/*#define DIVINT(n, m) (n >= 0 ? \
	(m > 0 ? (n + m / 2) / m : -((n + (-m) / 2) / (-m))) :\
	(m > 0 ? -((-n + m / 2) / m) : (n + (-m) / 2)) / (-m))
*/

#define DIVINT(n, m) s_divint(n, m)

static inline int
s_divint(int n, int m)
{
	if (n >= 0) {
		if (m > 0)
			return (n + m / 2) / m;
		else return -((n + (-m) / 2) / (-m));
	} else {
		n = -n;
		if (m > 0)
			return -(n + m / 2) / m;
		else return (n + (-m) / 2) / (-m);
	}
}

static inline void
s_pset_nocheck(int x, int y)
{
	*((my_exchange_buffer ? my_textscreen : my_graphscreen) + x + y * my_width) = s_gcolor;
	lm_redraw(x, y, 1, 1);
}

static void
s_pset(int x, int y)
{
	if (x >= 0 && x < my_width && y >= 0 && y < my_height) {
		s_pset_nocheck(x, y);
	} else {
	/*	fprintf(stderr, "pset out of range: (%d, %d)\n", x, y);  */
	}
}

static void
s_hline(int x1, int x2, int y, pixel_t col)
{
	pixel_t *p;
	int xx;
	if (x1 >= my_width || x2 < 0 || y < 0 || y >= my_height)
		return;
	if (x1 < 0)
		x1 = 0;
	if (x2 >= my_width)
		x2 = my_width - 1;
	p = (my_exchange_buffer ? my_textscreen : my_graphscreen) + x1 + y * my_width;
	xx = x1;
	while (xx++ <= x2)
		*p++ = col;
	lm_redraw(x1, y, x2 - x1 + 1, 1);
}

#define s_intersect_x(x1, y1, x2, y2, clip_x, w, wx, wy) \
	(x1 == x2 ? (x1 >= cx1 && x1 < cx2) : \
		(w = (y2 - y1) * (clip_x - x1) / (x2 - x1) + y1, \
		( (w < cy1 || w >= cy2) ? 0 : (wx = clip_x, wy = w, 1))))

#define s_intersect_y(x1, y1, x2, y2, clip_y, w, wx, wy) \
	(y1 == y2 ? (y1 >= cy1 && y1 < cy2) : \
		(w = (x2 - x1) * (clip_y - y1) / (y2 - y1) + x1, \
		( (w < cx1 || w >= cx2) ? 0 : (wx = w, wy = clip_y, 1))))

static void
s_line(int x1, int y1, int x2, int y2)
{
	int i, wx, wy, sx, sy;
	
	/*  Clipping  */
	int cx1 = 0;
	int cy1 = 0;
	int cx2 = my_width - 1;
	int cy2 = my_height - 1;
	
	int sx1 = x1, sx2 = x2, sy1 = y1, sy2 = y2;

	/*  Out-of-bounds flags  */
	u_int8_t f1 = 0;
	u_int8_t f2 = 0;
	if (x1 < cx1) f1 |= 1;
	if (x1 > cx2) f1 |= 2;
	if (y1 < cy1) f1 |= 4;
	if (y1 > cy2) f1 |= 8;
	if (x2 < cx1) f2 |= 1;
	if (x2 > cx2) f2 |= 2;
	if (y2 < cy1) f2 |= 4;
	if (y2 > cy2) f2 |= 8;
	
	if ((f1 & f2) != 0)
		return;  /*  Invisible  */
	
	if (f1 != 0) {
		/*  (x1, y1) needs clipping  */
		wx = x1;
		wy = y1;
		if (!((f1 & 1) && s_intersect_x(x1, y1, x2, y2, cx1, i, wx, wy)) &&
			!((f1 & 2) && s_intersect_x(x1, y1, x2, y2, cx2, i, wx, wy)) &&
			!((f1 & 4) && s_intersect_y(x1, y1, x2, y2, cy1, i, wx, wy)) &&
			!((f1 & 8) && s_intersect_y(x1, y1, x2, y2, cy2, i, wx, wy)))
			return;  /*  Invisible  */
		x1 = wx;
		y1 = wy;
	}
	if (f2 != 0) {
		/*  (x2, y2) needs clipping  */
		wx = x2;
		wy = y2;
		if (!((f2 & 1) && s_intersect_x(x1, y1, x2, y2, cx1, i, wx, wy)) &&
			!((f2 & 2) && s_intersect_x(x1, y1, x2, y2, cx2, i, wx, wy)) &&
			!((f2 & 4) && s_intersect_y(x1, y1, x2, y2, cy1, i, wx, wy)) &&
			!((f2 & 8) && s_intersect_y(x1, y1, x2, y2, cy2, i, wx, wy)))
			return;  /*  Invisible  */
		x2 = wx;
		y2 = wy;
	}
	
	wx = (x2 > x1 ? x2 - x1 : x1 - x2);
	wy = (y2 > y1 ? y2 - y1 : y1 - y2);
	sx = (x2 > x1 ? 1 : -1);
	sy = (y2 > y1 ? 1 : -1);

	if (0) {
		double fabs(double);
		if (wx > wy) {
			double dr1, dr2, dy1, dy2;
			dr1 = ((double)(sy2 - sy1) / (double)(sx2 - sx1));
			dr2 = sy1 - dr1 * sx1;
			dy1 = dr1 * x1 + dr2;
			dy2 = dr1 * x2 + dr2;
			if (fabs(dy1 - y1) >= 1.0 || fabs(dy2 - y2) >= 1.0) {
				printf("clipping NG: (%d,%d)-(%d,%d)/(%d,%d)-(%d,%d)\n", sx1, sy1, sx2, sy2, x1, y1, x2, y2);
			}
		} else {
			double dr1, dr2, dx1, dx2;
			dr1 = ((double)(sx2 - sx1) / (double)(sy2 - sy1));
			dr2 = sx1 - dr1 * sy1;
			dx1 = dr1 * y1 + dr2;
			dx2 = dr1 * y2 + dr2;
			if (fabs(dx1 - x1) >= 1.0 || fabs(dx2 - x2) >= 1.0) {
				printf("clipping NG: (%d,%d)-(%d,%d)/(%d,%d)-(%d,%d)\n", sx1, sy1, sx2, sy2, x1, y1, x2, y2);
			}
		}
	}
	
	if (wx > wy) {
		int e = -wx;
		for (i = 0; i <= wx; i++) {
			s_pset_nocheck(x1, y1);
			x1 += sx;
			e += 2 * wy;
			if (e >= 0) {
				y1 += sy;
				e -= 2 * wx;
			}
		}
	} else {
		int e = -wy;
		for (i = 0; i <= wy; i++) {
			s_pset_nocheck(x1, y1);
			y1 += sy;
			e += 2 * wx;
			if (e >= 0) {
				x1 += sx;
				e -= 2 * wy;
			}
		}
	}
}

#define PI 3.141592653589793

static void
s_arc(int ox, int oy, int rad, double start, double end)
{
	int x, y, f, st, wd, d, d8;
	double ct, tn;
	if (rad == 0) {
		s_pset(ox, oy);
		return;
	} else if (rad < 0)
		return;
	d = ceil(0.7071067811865476 * rad);
	if (end - start == 360.0) {
		start = 0.0;
		end = 360.0;
	}
	if (start == 0.0) {
		st = 0;
	} else {
		start -= floor(start / 360.0) * 360.0;
		start *= PI / 180.0;
		ct = -cos(start) / sin(start);
		tn = tan(start);
		if (fabs(tn) <= fabs(ct)) {
			tn += (start >= PI * 1.74 ? 8 : (start >= PI * 0.74 && start <= PI * 1.26 ? 4 : 0));
		} else {
			tn = ct + (start >= PI * 0.24 && start <= PI * 0.76 ? 2 : 6);
		}
		st = floor(tn * d);
	}
	if (end == 360.0) {
		wd = 8 * d - st;
	} else {
		end -= floor(start / 360.0) * 360.0;
		end *= PI / 180.0;
		ct = -cos(end) / sin(end);
		tn = tan(end);
		if (fabs(tn) <= fabs(ct)) {
			tn += (end >= PI * 1.74 ? 8 : (end >= PI * 0.74 && end <= PI * 1.26 ? 4 : 0));
		} else {
			tn = ct + (end >= PI * 0.24 && end <= PI * 0.76 ? 2 : 6);
		}
		wd = ceil(tn * d) - st;
	}
	d8 = d * 8;
	if (wd < 0)
		wd += d8;
	st -= d8;

	x = rad;
	y = 0;
	f = 3 - 2 * rad;
	while (x >= y) {
		if ((y - st) % d8 <= wd)
			s_pset(ox + x, oy + y);
		if ((2 * d - y - st) % d8 <= wd)
			s_pset(ox + y, oy + x);
		if ((2 * d + y - st) % d8 <= wd)
			s_pset(ox - y, oy + x);
		if ((4 * d - y - st) % d8 <= wd)
			s_pset(ox - x, oy + y);
		if ((4 * d + y - st) % d8 <= wd)
			s_pset(ox - x, oy - y);
		if ((6 * d - y - st) % d8 <= wd)
			s_pset(ox - y, oy - x);
		if ((6 * d + y - st) % d8 <= wd)
			s_pset(ox + y, oy - x);
		if ((8 * d - y - st) % d8 <= wd)
			s_pset(ox + x, oy - y);
		if (f >= 0) {
			x--;
			f -= 4 * x;
		}
		y++;
		f += 4 * y + 2;
	}
}

static void
s_obarc(int ox, int oy, int rx, int ry, double start, double end, int drawflag)
{
	int a, b, x, y, f, h, d, dx, dy, st, wd, d8, dd;
	int px, py, qx, qy, fy0, fy1, fx;
	u_int8_t wf, ff, fill, stroke;
	double ct, tn;
	
	if (rx < 0 || ry < 0)
		return;
	stroke = drawflag & 1;
	fill = drawflag & 2;
	if (rx == 0) {
		if (fill || stroke) {
			if (ry == 0)
				s_pset(ox, oy);
			else
				s_line(ox, oy - ry, ox, oy + ry);
		}
		return;
	} else if (ry == 0) {
		if (fill || stroke)
			s_line(ox - rx, oy, ox + rx, oy);
		return;
	}
	
	dx = ceil(0.7071067811865476 * rx);
	dy = ceil(0.7071067811865476 * ry);
	d = (dx > dy ? dx : dy);
	
	if (end - start == 360.0) {
		start = 0.0;
		end = 360.0;
	}
	if (start == 0.0) {
		st = 0;
	} else {
		start -= floor(start / 360.0) * 360.0;
		start *= PI / 180.0;
		ct = 1.41421356 * cos(start);
		tn = 1.41421356 * sin(start);
		if (fabs(tn) <= fabs(ct)) {
			if (start >= PI * 1.74)
				tn += 8;
			else if (start >= PI * 0.74 && start <= PI * 1.26)
				tn = 4 - tn;
		} else {
			if (start >= PI * 0.24 && start <= PI * 0.76)
				tn = 2 - ct;
			else tn = 6 + ct;
		}
		st = floor(tn * d + 0.5);
	}
	if (end == 360.0) {
		wd = 8 * d - st;
		end = 0.0;
	} else {
		end -= floor(start / 360.0) * 360.0;
		end *= PI / 180.0;
		ct = 1.41421356 * cos(end);
		tn = 1.41421356 * sin(end);
		if (fabs(tn) <= fabs(ct)) {
			if (end >= PI * 1.74)
				tn += 8;
			else if (end >= PI * 0.74 && end <= PI * 1.26)
				tn = 4 - tn;
		} else {
			if (end >= PI * 0.24 && end <= PI * 0.76)
				tn = 2 - ct;
			else tn = 6 + ct;
		}
		wd = floor(tn * d + 0.5) - st;
	}
	d8 = d * 8;
	if (wd < 0)
		wd += d8;
	st -= d8;
	qx = floor(rx * cos(end) + 0.5);
	qy = floor(ry * sin(end) + 0.5);
	px = floor(rx * cos(start) + 0.5) - qx;
	py = floor(ry * sin(start) + 0.5) - qy;
	a = ry * ry;
	b = rx * rx;
	x = rx;
	y = 0;
	f = -2 * rx * a + a + 2 * b;
	h = -4 * rx * a + 2 * a + b;
	fy0 = fy1 = -1;
	if (py < 0) {
		px = -px;
		py = -py;
	}
	while (x >= 0) {
		wf = 0;
		if (x >= dx || y < dy) {
			dd = (y * d + dy / 2) / dy;
			if ((dd - st) % d8 <= wd)
				wf |= 1;
			if ((4 * d - dd - st) % d8 <= wd)
				wf |= 2;
			if ((4 * d + dd - st) % d8 <= wd)
				wf |= 4;
			if ((8 * d - dd - st) % d8 <= wd)
				wf |= 8;
		} else {
			dd = (x * d + dx / 2) / dx;
			if ((2 * d - dd - st) % d8 <= wd)
				wf |= 1;
			if ((2 * d + dd - st) % d8 <= wd)
				wf |= 2;
			if ((6 * d - dd - st) % d8 <= wd)
				wf |= 4;
			if ((6 * d + dd - st) % d8 <= wd)
				wf |= 8;
		}
		ff = 0;
		if (wf & 1) {
			if (fill && fy0 != y) {
				ff = 1;
				if (py == 0 || wf & 2)
					fx = -x;
				else {
					fx = qx + DIVINT(px * (y - qy), py);
					if (wf & 2)
						fx = -x;
					if (fx < -x)
						fx = -x;
					else if (fx > x)
						fx = x;
				}
				s_hline(ox + fx, ox + x, oy + y, s_fillcolor);
			}
			if (stroke)
				s_pset(ox + x, oy + y);
		}
		if (wf & 2) {
			if (fill && fy0 != y && ff == 0) {
				ff = 1;
				if (py == 0)
					fx = x;
				else {
					fx = qx + DIVINT(px * (y - qy), py);
					if (fx > x)
						fx = x;
					else if (fx < -x)
						fx = -x;
				}
				s_hline(ox - x, ox + fx, oy + y, s_fillcolor);
			}
			if (stroke)
				s_pset(ox - x, oy + y);
		}
		if (ff)
			fy0 = y;
		ff = 0;
		if (wf & 8) {
			if (fill && fy1 != y) {
				ff = 1;
				if (py == 0 || wf & 4)
					fx = -x;
				else {
					fx = qx + DIVINT(px * (-y - qy), py);
					if (fx < -x)
						fx = -x;
					else if (fx > x)
						fx = x;
				}
				s_hline(ox + fx, ox + x, oy - y, s_fillcolor);
			}
			if (stroke)
				s_pset(ox + x, oy - y);
		}
		if (wf & 4) {
			if (fill && fy1 != y && ff == 0) {
				ff = 1;
				if (py != 0) {
					fx = qx + DIVINT(px * (-y - qy), py);
					if (fx > x)
						fx = x;
					else if (fx < -x)
						fx = -x;
				} else fx = x;
				s_hline(ox - x, ox + fx, oy - y, s_fillcolor);
			}
			if (stroke)
				s_pset(ox - x, oy - y);
		}
		if (ff)
			fy1 = y;
		if (f >= 0) {
			x--;
			f -= 4 * a * x;
			h -= a * (4 * x - 2);
		}
		if (h < 0) {
			y++;
			f += b * (4 * y + 2);
			h += 4 * b * y;
		}
	}
}

static void
s_polyfill(int16_t *pts, int npts)
{
	int x1, y1, x2, y2, y3, dy1, dy2, dd, i, j, n;
	int16_t *nodes;
	u_int8_t *nodeflags;
	nodes = alloca((sizeof(int16_t) * 2 + sizeof(u_int8_t)) * npts);
	if (nodes == NULL) {
		fprintf(stderr, "Cannot allocate memory for drawing polylines\n");
		return;
	}
	nodeflags = (u_int8_t *)(nodes + npts * 2);
	memset(nodeflags, 0, npts);
	y2 = pts[npts * 2 - 1];
	y1 = pts[1];
	dy1 = y1 - y2;
	for (i = 0; i <= npts; i++) {
		/*  The first point is processed twice. Otherwise, the algorithm may
		    fail when initial dy1 == 0.  */
		if (i < npts - 1)
			y2 = pts[i * 2 + 3];
		else y2 = pts[(i + 1 - npts) * 2 + 1];
		dy2 = y2 - y1;
		if (dy2 == 0) {
			nodeflags[i] = 2;  /*  This edge is horizontal  */
		} else {
			if ((dy1 > 0 && dy2 > 0) || (dy1 < 0 && dy2 < 0))
				nodeflags[i % npts] = 1;  /*  This edge is not summit nor valley  */
			dy1 = dy2;  /*  dy1 is the 'last slope direction'  */
		}
		y1 = y2;
	}
	y1 = -10000000;
	y2 = 10000000;
	for (i = 0; i < npts; i++) {
		y3 = pts[i * 2 + 1];
		if (y1 < y3)
			y1 = y3;
		if (y2 > y3)
			y2 = y3;
	}
	for (y3 = y2; y3 <= y1; y3++) {
		if (y3 < 0 || y3 >= my_height)
			continue;
		n = 0;
		for (i = 0; i < npts; i++) {
			x1 = pts[i * 2];
			dy1 = pts[i * 2 + 1];
			if (i == npts - 1) {
				x2 = pts[0];
				dy2 = pts[1];
			} else {
				x2 = pts[i * 2 + 2];
				dy2 = pts[i * 2 + 3];
			}
			dy1 -= y3;
			dy2 -= y3;
			dd = dy2 - dy1;
			if ((dy1 > 0 && dy2 > 0) || (dy1 < 0 && dy2 < 0))
				continue;
			if (dy1 == 0 && nodeflags[i])
				continue;
			if ((dy1 > 0 && dy1 + dy2 >= 0) || (dy1 < 0 && dy1 + dy2 <= 0)) {
				x1 = x1 - DIVINT(dy1 * (x2 - x1), dd);
			} else {
				x1 = x2 - DIVINT(dy2 * (x2 - x1), dd);
			}
			for (j = 0; j < n; j++) {
				if (nodes[j] >= x1)
					break;
			}
			if (j < n)
				memmove(nodes + j + 1, nodes + j, sizeof(nodes[0]) * (n - j));
			nodes[j] = x1;
			n++;
		}
		if (n % 2 != 0) {
			fprintf(stderr, "number of nodes is odd\n");
			continue;
		}
		for (i = 0; i < n; i += 2)
			s_hline(nodes[i], nodes[i + 1], y3, s_fillcolor);
	}
}

static void
s_poly(int16_t *pts, int npts, int drawflag)
{
	int i;
	if (drawflag & 2)
		s_polyfill(pts, npts);
	if (drawflag & 1) {
		for (i = 0; i < npts - 1; i++)
			s_line(pts[i * 2], pts[i * 2 + 1], pts[i * 2 + 2], pts[i * 2 + 3]);
		s_line(pts[i * 2], pts[i * 2 + 1], pts[0], pts[1]);
	}
}

static void
s_box(int x, int y, int width, int height, int drawflag)
{
	int16_t pts[10];
	if (width <= 0 || height <= 0)
		return;
	pts[0] = pts[6] = pts[8] = x;
	pts[1] = pts[3] = pts[9] = y;
	pts[2] = pts[4] = x + width;
	pts[5] = pts[7] = y + height;
	s_poly(pts, 5, drawflag);
}

static void
s_roundbox(int x, int y, int width, int height, int rx, int ry, int drawflag)
{
	int16_t pts[18];
	if (width <= 0 || height <= 0)
		return;
	if (rx > width / 2)
		rx = width / 2;
	if (ry > height / 2)
		ry = height / 2;
	if (rx <= 0 || ry <= 0)
		s_box(x, y, width, height, drawflag);

	pts[0] = pts[10] = pts[16] = x + rx;
	pts[2] = pts[8] = x + width - rx;
	pts[4] = pts[6] = x + width;
	pts[12] = pts[14] = x;
	pts[1] = pts[3] = pts[17] = y;
	pts[5] = pts[15] = y + ry;
	pts[7] = pts[13] = y + height - ry;
	pts[9] = pts[11] = y + height;
	
	if (drawflag & 2)
		s_polyfill(pts, 9);

	s_obarc(pts[0], pts[5], rx, ry, 180.0, 270.0, drawflag);
	s_obarc(pts[2], pts[5], rx, ry, 270.0, 0.0, drawflag);
	s_obarc(pts[2], pts[7], rx, ry, 0.0, 90.0, drawflag);
	s_obarc(pts[0], pts[7], rx, ry, 90.0, 180.0, drawflag);
	
	if (drawflag & 1) {
		s_line(pts[0], pts[1], pts[2], pts[3]);
		s_line(pts[4], pts[5], pts[6], pts[7]);
		s_line(pts[8], pts[9], pts[10], pts[11]);
		s_line(pts[12], pts[13], pts[14], pts[15]);
	}
}

int
l_gcolor(lua_State *L)
{
	if (lua_gettop(L) == 0 || lua_toboolean(L, 1) == 0)
		s_strokeflag = 0;
	else {
		s_strokeflag = 1;
		if (!lua_isboolean(L, 1))
			s_gcolor = luaL_checkinteger(L, 1);
	}
	return 0;
}

int
l_fillcolor(lua_State *L)
{
	if (lua_gettop(L) == 0 || lua_toboolean(L, 1) == 0)
		s_fillflag = 0;
	else {
		s_fillflag = 2;
		if (!lua_isboolean(L, 1))
			s_fillcolor = luaL_checkinteger(L, 1);
	}
	return 0;
}

int
l_pset(lua_State *L)
{
	int x = LUA_AUTOINT(L, 1);
	int y = LUA_AUTOINT(L, 2);
	if (x >= 0 && x < my_width && y >= 0 && y < my_height)
		s_pset(x, y);
	return 0;
}

int
l_line(lua_State *L)
{
	int x1 = LUA_AUTOINT(L, 1);
	int y1 = LUA_AUTOINT(L, 2);
	int x2 = LUA_AUTOINT(L, 3);
	int y2 = LUA_AUTOINT(L, 4);
	s_line(x1, y1, x2, y2);
	return 0;
}

int
l_circle(lua_State *L)
{
	int nargs = lua_gettop(L);
	int ox = LUA_AUTOINT(L, 1);
	int oy = LUA_AUTOINT(L, 2);
	int rad1 = LUA_AUTOINT(L, 3);
	int rad2 = rad1;
	if (nargs >= 4)
		rad2 = LUA_AUTOINT(L, 4);
	s_obarc(ox, oy, rad1, rad2, 0.0, 360.0, s_strokeflag | s_fillflag);
	return 0;
}

int
l_arc(lua_State *L)
{
	int nargs = lua_gettop(L);
	int ox = LUA_AUTOINT(L, 1);
	int oy = LUA_AUTOINT(L, 2);
	double start = luaL_checknumber(L, 3);
	double end = luaL_checknumber(L, 4);
	int rad1 = LUA_AUTOINT(L, 5);
	int rad2 = rad1;
	if (nargs >= 6)
		rad2 = LUA_AUTOINT(L, 6);
	s_obarc(ox, oy, rad1, rad2, start, end, s_strokeflag | s_fillflag);
	return 0;
}

int
l_fan(lua_State *L)
{
	int nargs = lua_gettop(L);
	int ox = LUA_AUTOINT(L, 1);
	int oy = LUA_AUTOINT(L, 2);
	double start = luaL_checknumber(L, 3);
	double end = luaL_checknumber(L, 4);
	double d;
	int rad1 = LUA_AUTOINT(L, 5);
	int rad2 = rad1;
	int16_t p[8];
	if (nargs >= 6)
		rad2 = LUA_AUTOINT(L, 6);
	if (start != 0.0 || end != 360.0 || s_fillflag) {
		d = end - start;
		d -= floor(d / 360.0) * 360.0;
		if (d > 180.0) {
			/*  Draw and fill two arcs separately  */
			d = start + 180.0;
			d -= floor(d / 360.0) * 360.0;
		} else d = start;
	}
	p[0] = ox;
	p[1] = oy;
	p[2] = ox + floor(rad1 * cos(d * (PI / 180.0)) + 0.5);
	p[3] = oy + floor(rad2 * sin(d * (PI / 180.0)) + 0.5);
	p[4] = ox + floor(rad1 * cos(end * (PI / 180.0)) + 0.5);
	p[5] = oy + floor(rad2 * sin(end * (PI / 180.0)) + 0.5);
	if (s_fillflag) {
		s_polyfill(p, 3);
	}
	if (d != start)
		s_obarc(ox, oy, rad1, rad2, start, d, s_strokeflag | s_fillflag);
	s_obarc(ox, oy, rad1, rad2, d, end, s_strokeflag | s_fillflag);		
	if (s_strokeflag) {
		s_line(ox, oy, ox + floor(rad1 * cos(start * (PI / 180.0)) + 0.5), oy + floor(rad2 * sin(start * (PI / 180.0)) + 0.5));
		s_line(ox, oy, p[4], p[5]);
	}
	return 0;	
}

int
l_poly(lua_State *L)
{
	int16_t *buf;
	int i;
	if (lua_istable(L, 1)) {
		int n = lua_rawlen(L, 1);
		if (n % 2 != 0) {
			fprintf(stderr, "The y-coordinate of the last point is missing\n");
			return 0;
		}
		buf = (int16_t *)alloca(sizeof(int16_t) * n);
		if (buf == 0) {
			fprintf(stderr, "Cannot allocate buffer for drawing polygon\n");
			return 0;
		}
		for (i = 0; i < n; i++) {
			lua_rawgeti(L, 1, i + 1);
			buf[i] = LUA_AUTOINT(L, 2);
			lua_pop(L, 1);
		}
		s_poly(buf, n / 2, s_strokeflag | s_fillflag);
	}
	return 0;
}

int
l_box(lua_State *L)
{
	int x1 = LUA_AUTOINT(L, 1);
	int y1 = LUA_AUTOINT(L, 2);
	int wid = LUA_AUTOINT(L, 3);
	int hei = LUA_AUTOINT(L, 4);
	s_box(x1, y1, wid, hei, s_strokeflag | s_fillflag);
	return 0;
}

int
l_rbox(lua_State *L)
{
	int x1 = LUA_AUTOINT(L, 1);
	int y1 = LUA_AUTOINT(L, 2);
	int wid = LUA_AUTOINT(L, 3);
	int hei = LUA_AUTOINT(L, 4);
	int rx = LUA_AUTOINT(L, 5);
	int ry = LUA_AUTOINT(L, 6);
	s_roundbox(x1, y1, wid, hei, rx, ry, s_strokeflag | s_fillflag);
	return 0;
}

int
l_gcls(lua_State *L)
{
	memset(my_graphscreen, 0, my_width * my_height * sizeof(u_int16_t));
	lm_redraw(0, 0, my_width, my_height);
	return 0;
}

/*  gmode(0): normal, gmode(1): 2x2 expanded screen  */
int
l_gmode(lua_State *L)
{
	int mode = LUA_AUTOINT(L, 1);
	my_graphic_mode = mode;
	if (mode == 1) {
		my_max_x = my_width / 16;
		my_max_y = my_height / 32;
		if (my_cursor_x >= my_max_x) {
			my_cursor_y++;
			my_cursor_x = 0;
		}
		if (my_cursor_y >= my_max_y)
			my_cursor_y = my_max_y - 1;
	} else {
		my_max_x = my_width / 8;
		my_max_y = my_height / 16;
	}
	lm_redraw(0, 0, my_width, my_height);
	return 0;
}

/*  patdef(num, width, height, table)
    Define a pattern #num (num in 1..256). Table is an array of color codes (column first).  */
int
l_patdef(lua_State *L)
{
	int width, height, len, i, num, len2;
	pixel_t *p;
	num = luaL_checkinteger(L, 1);
	width = luaL_checkinteger(L, 2);
	height = luaL_checkinteger(L, 3);
	luaL_argcheck(L, num >= 1 && num <= 256, 1, "must be in 1..256");
	luaL_argcheck(L, width > 0, 2, "must be a positive integer");
	luaL_argcheck(L, height > 0, 3, "must be a positive integer");
	luaL_argcheck(L, lua_istable(L, 4), 4, "table expected");
	len = width * height;
	num--;
	if (my_patbuffer_count + len > my_patbuffer_size)
		return luaL_error(L, "out of pattern memory; pattern size (%dx%d) is too large", width, height);

	len2 = my_patterns[num].width * my_patterns[num].height;
	if (len2 == 0) {
		/*  New entry  */
		if (num == 0)
			my_patterns[num].ofs = 0;
		else {
			LPattern *lp = my_patterns + num - 1;
			(lp + 1)->ofs = lp->ofs + lp->width * lp->height;
		}
	}		
	/*  Move the already existing data  */
	p = my_patbuffer + my_patterns[num].ofs;
	
	memmove(my_patbuffer + len, my_patbuffer + len2, sizeof(u_int16_t) * (my_patbuffer_size - (len > len2 ? len : len2)));
	for (i = num + 1; i < NUMBER_OF_PATTERNS; i++)
		my_patterns[i].ofs += len - len2;

	for (i = 0; i < len; i++) {
		lua_rawgeti(L, 4, i + 1);
		*p++ = luaL_checkinteger(L, -1);
		lua_pop(L, 1);
	}
	my_patterns[num].width = width;
	my_patterns[num].height = height;
	
	return 0;
}

/*  patundef(num)
  Forget the pre-defined pattern #num.  */
int
l_patundef(lua_State *L)
{
	int i, len2;
	pixel_t *p;
	int num = luaL_checkinteger(L, 1);
	luaL_argcheck(L, num >= 1 && num <= 256, 1, "must be in 1..256");
	num--;
	/*  Move the already existing data  */
	p = my_patbuffer + my_patterns[num].ofs;
	len2 = my_patterns[num].width * my_patterns[num].height;
	memmove(my_patbuffer, my_patbuffer + len2, sizeof(pixel_t) * (my_patbuffer_size - len2));
	for (i = num + 1; i < NUMBER_OF_PATTERNS; i++)
		my_patterns[i].ofs -= len2;
	my_patterns[num].width = my_patterns[num].height = my_patterns[num].ofs = 0;
	return 0;
}

static void
s_patwrite(int n, int x1, int y1, u_int32_t mask, int eraseflag)
{
	int x2, y2;
	int dx, dy, wx, wy, ww;
	int x, y, r, g, b;
	pixel_t *p1, *p2;

	wx = my_patterns[n].width;
	wy = my_patterns[n].height;
	x2 = x1 + wx;
	y2 = y1 + wy;
	if (x2 < 0 || x1 >= my_width || y2 < 0 || y1 >= my_height)
		return;
	dx = dy = ww = 0;
	if (x1 < 0) {
		dx = -x1;
		wx -= dx;
		x1 = 0;
	}
	if (y1 < 0) {
		dy = -y1;
		wy -= dy;
		y1 = 0;
	}
	if (x2 > my_width) {
		x2 = my_width;
		wx = x2 - x1;
	}
	if (y2 > my_height) {
		y2 = my_height;
		wy = y2 - y1;
	}
	ww = my_patterns[n].width - wx;
	p1 = (my_exchange_buffer ? my_textscreen : my_graphscreen) + y1 * my_width + x1;
	p2 = my_patbuffer + my_patterns[n].ofs + dy * my_patterns[n].width + dx;
	if (eraseflag == 0) {
		for (y = y1; y < y2; y++) {
			for (x = x1; x < x2; x++) {
				u_int32_t col = *p2;
				if (col != 0) {
#if BYTES_PER_PIXEL == 2
					r = (((col >> 11) & 0x1f) * ((mask >> 11) & 0x1f)) / 0x1f;
					g = (((col >> 5) & 0x3f) * ((mask >> 5) & 0x3f)) / 0x3f;
					b = ((col & 0x1f) * (mask & 0x1f)) / 0x1f;
#else
					r = (((col >> 16) & 0xff) * ((mask >> 16) & 0xff)) / 0xff;
					g = (((col >> 8) & 0xff) * ((mask >> 8) & 0xff)) / 0xff;
					b = ((col & 0xff) * (mask & 0xff)) / 0xff;
#endif
					*p1 = col;
				}
				p1++;
				p2++;
			}
			p1 += my_width - wx;
			p2 += ww;
		}	
	} else {
		for (y = y1; y < y2; y++) {
			for (x = x1; x < x2; x++) {
				if (*p2 != 0)
					*p1 = 0;
				p1++;
				p2++;
			}
			p1 += my_width - wx;
			p2 += ww;
		}	
	}
	lm_redraw(x1, y1, wx, wy);
}

/*  patdraw(n, x, y [, mask])
    Draw the defined pattern (#n) at position (x, y).
    Mask is a color mask, where the original RGB components are attenuated by this RGB values. */
int
l_patdraw(lua_State *L)
{
	int n, x1, y1;
	u_int32_t mask;
	n = luaL_checkinteger(L, 1);
	if (n <= 0 || n > NUMBER_OF_PATTERNS)
		return luaL_argerror(L, 1, "pattern number must be in (1..256)");
	n--;
	if (my_patterns[n].width == 0)
		return luaL_error(L, "pattern (%d) is undefined", n + 1);
	x1 = luaL_checkinteger(L, 2);
	y1 = luaL_checkinteger(L, 3);
	if (lua_gettop(L) >= 4)
		mask = luaL_checkinteger(L, 4);
	else mask = (BYTES_PER_PIXEL == 2 ? 0xffff : 0xffffffff);
	s_patwrite(n, x1, y1, mask, 0);
	return 0;
}

/*  paterase(n, x, y)
 Erase the defined pattern (#n) at position (x, y).  */
int
l_paterase(lua_State *L)
{
	int n, x1, y1;
	n = luaL_checkinteger(L, 1);
	if (n <= 0 || n > NUMBER_OF_PATTERNS)
		return luaL_argerror(L, 1, "pattern number must be in (1..256)");
	n--;
	if (my_patterns[n].width == 0)
		return luaL_error(L, "pattern (%d) is undefined", n + 1);
	x1 = luaL_checkinteger(L, 2);
	y1 = luaL_checkinteger(L, 3);
	s_patwrite(n, x1, y1, 0, 1);
	return 0;
}

/*  rgb(r, g, b[, a])
 Return a color code. The arguments are floating number (0.0 to 1.0).  */
int
l_rgb(lua_State *L)
{
	int n;
	double r, g, b, a;
	n = lua_gettop(L);
	if (n >= 4)
		a = luaL_checknumber(L, 4);
	else a = 1.0;
	r = luaL_checknumber(L, 1);
	g = luaL_checknumber(L, 2);
	b = luaL_checknumber(L, 3);
	lua_pushinteger(L, RGBAFLOAT(r, g, b, a));
	return 1;
}

