/*
 *  luacon.c
 *  LuaConsole
 *
 *  Created by 永田 央 on 16/07/14.
 *  Copyright 2016 __MyCompanyName__. All rights reserved.
 *
 */

#include "lua.h"
#include "lualib.h"
#include "lauxlib.h"
#include "screen.h"

#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>

extern int lm_load_file(const char *fname);
extern int lm_list(int sline, int eline);

u_int8_t gSourceTouched;

static int
l_list(lua_State *L)
{
	int st, en;
	int nargs = lua_gettop(L);
	st = 1;
	en = 999999;
	if (nargs > 0) {
		st = en = luaL_checkinteger(L, 1);
		if (nargs > 1) {
			en = luaL_checkinteger(L, 2);
		} else en += 9;
	}
	lm_list(st, en);
	return 0;
}

static int
l_edit(lua_State *L)
{
	extern int lm_edit(void);	
	gSourceTouched = 1;
	lm_edit();
	return 0;
}

static int
l_load(lua_State *L)
{
	const char *p;
	size_t len;
	p = luaL_checklstring(L, -1, &len);
	if (p != NULL)
		lm_load_file(p);
	gSourceTouched = 0;
	return 0;
}

static int
l_wait(lua_State *L)
{
	double d = luaL_checknumber(L, 1);
	usleep(d * 1000);
	return 0;
}

static const struct luaL_Reg sys_lib[] = {
	{"load", l_load},
	{"edit", l_edit},
	{"list", l_list},
	{"wait", l_wait},
	{NULL, NULL}
};

int
luaopen_sys(lua_State *L) 
{
#if LUA_VERSION_NUM >= 502
	luaL_newlib(L, sys_lib);
#else
	luaL_register(L, "sys", sys_lib);
#endif
	return 1;
}


static lua_State *
s_lua_init(void)
{
	lua_State *L = luaL_newstate();
	luaL_openlibs(L);
	luaL_requiref(L, "con", luaopen_con, 1);
	luaL_requiref(L, "sys", luaopen_sys, 1);
	gSourceLimit = 65536;
	gSourceBasePtr = (u_int8_t *)realloc(gSourceBasePtr, gSourceLimit);
	gSourceEnd = 0;
	
	return L;
}

int
lua_main(void)
{
	int n, n1, n2, n3, status;
	const char *version_str;
	FILE *fp;

	lua_State *L = s_lua_init();
	
	lua_getglobal(L, "_VERSION");
	version_str = lua_tostring(L, -1);

	lm_puts("Lua Console [るあこん] v0.1\n");
	lm_puts(version_str);
	lm_puts(", Copyright © 1994-2006 Lua.org, PUC-Rio\n");

	if (chdir("lua") != 0) {
		lm_puts("The directory 'lua' was not found");
		if (mkdir("lua", 0755) == 0) {
			lm_puts(". Created one.\n");
			chdir("lua");
		} else {
			lm_puts(", and cannot be created.\n");
		}
	}
	
	status = 0;
	fp = fopen("../startup.lua", "r");
	if (fp != NULL) {
		fclose(fp);
		lm_puts("Running startup.lua...\n");
		status = luaL_dofile(L, "../startup.lua");
	} else {
		lm_puts("startup.lua was not found.\n");
	}
	if (status == 0)
		lm_puts("Ready.\n");

	gSourceTouched = 0;
	
	while (1) {
		static char s[1024];

		if (status != 0) {
			/*  The error message should be on the top of Lua stack  */
			size_t len;
			const char *p;
			p = luaL_checklstring(L, -1, &len);
			if (strncmp(p, "[string \"\"]:", 12) == 0) {
				p += 12;
				lm_puts("Line ");
			} else if (strncmp(p, "[string \"S\"]:1", 14) == 0) {
				p += 14;
				lm_puts("Line 0");
			}
			lm_puts(p);
			status = 0;
			lm_puts("\n");
		}
		n = lua_gettop(L);
		lua_pop(L, n);
		
		lm_puts(">> ");
		if (lm_getline(s + 4, sizeof s - 6) > 0) {
			/*  Offset by 4, for prefix 'sys.'  */
			for (n = 4; s[n] > 0 && s[n] <= ' '; n++);
			if (s[n] == 0)
				continue;
			if (isalpha(s[n])) {
				for (n1 = n + 1; isalnum(s[n1]); n1++);
				n1 -= n;
				if (strncmp(s + n, "new", n1) == 0) {
					/*  Initialize the program and lua engine  */
					if (gSourceEnd > 0) {
						lm_puts("The program will be deleted. Do you really want to do this? (y/N) ");
						while (lm_getline(s, sizeof s) < 1);
						if (s[0] != 'y' && s[0] != 'Y')
							continue;
					}
					gSourceEnd = 0;
					gSourceBasePtr[0] = 0;
					lua_close(L);
					L = s_lua_init();
					status = 0;
					gSourceTouched = 0;
					continue;
				}
				if (strncmp(s + n, "exit", n1) == 0)
					break;
				if (strncmp(s + n, "run", n1) == 0) {
					if (gSourceEnd == 0)
						continue;
					status = (luaL_loadbuffer(L, (char *)gSourceBasePtr, gSourceEnd, "")
							  || lua_pcall(L, 0, LUA_MULTRET, 0));
					continue;
				}
				if (strncmp(s + n, "load", n1) == 0 ||
					strncmp(s + n, "edit", n1) == 0 ||
					strncmp(s + n, "list", n1) == 0) {
					/*  Insert '()' if not present  */
					n3 = n + strlen(s + n);
					while (s[n3 - 1] <= ' ')
						n3--; /*  strip whitespace at the end  */
					for (n2 = n + n1; n2 <= n3; n2++) {
						if (s[n2] == '(') {
							/*  '(' is already present  */
							break;
						} else if (s[n2] > ' ' || n2 == n3) {
							/*  No '()': we need to insert them  */
							memmove(s + n2 + 1, s + n2, n3 - n2);
							s[n2] = '(';
							s[n3 + 1] = ')';
							s[n3 + 2] = 0;
							break;
						}
					}
					strncpy(s + n - 4, "sys.", 4);
					n -= 4;
				}
			}
			status = (luaL_loadbuffer(L, s + n, strlen(s + n), "S")
					  || lua_pcall(L, 0, LUA_MULTRET, 0));
		}
	}
	lm_puts("[LuaCon Finished]");
	return n;
}
