/*
 *  raspi_main.c
 *  LuaConsole
 *
 *  Created by Toshi Nagata on 2016/02/07.
 *  Copyright 2016 Toshi Nagata.  All rights reserved.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#ifndef __BAREMETAL__

#include <fcntl.h>
#include <termios.h>

#endif

#if 0
#pragma mark ====== Framebuffer ======
#endif

#ifndef __CONSOLE__
#if __BAREMETAL__
#include <fb.h>
#else
#include <linux/fb.h>
#include <sys/mman.h>
#endif
#endif

#include "screen.h"


static int s_original_bpp;

static char *s_framebuffer;
static int s_fbfd = 0;

static void
s_pb_raspi_framebuffer(void)
{
#ifndef __CONSOLE__
#if __BAREMETAL__
	fb_init(0, 0);
	my_width = fb_width;
	my_height = fb_height;
	s_framebuffer = (char *)fb_addr;
#else
	struct fb_var_screeninfo vinfo;
	struct fb_fix_screeninfo finfo;
	
	/*  Open frame buffer device  */
	s_fbfd = open("/dev/fb0", O_RDWR);
	if (!s_fbfd) {
		printf("Error: cannot open framebuffer device.\n");
		return;
	}
	if (ioctl(s_fbfd, FBIOGET_VSCREENINFO, &vinfo)) {
		printf("Error reading variable information.\n");
		return;
	}
	printf("Frame buffer %dx%d, %dbpp\n", vinfo.xres, vinfo.yres, vinfo.bits_per_pixel );
	
	s_original_bpp = vinfo.bits_per_pixel;
	vinfo.bits_per_pixel = 16;
	if (ioctl(s_fbfd, FBIOPUT_VSCREENINFO, &vinfo)) {
		printf("Error setting variable information.\n");
		return;
	}
	
	my_width = vinfo.xres;
	my_height = vinfo.yres;
	
	if (ioctl(s_fbfd, FBIOGET_FSCREENINFO, &finfo)) {
		printf("Error reading fixed information.\n");
		return;
	}
	
	/*  Map the memory buffer  */
	s_framebuffer = (char *)mmap(0, finfo.smem_len, PROT_READ | PROT_WRITE, MAP_SHARED, s_fbfd, 0);
	if (s_framebuffer == (char *)-1) {
		printf("Failed to mmap.\n");
		return;
	}
	printf("mmap'ed buffer %p\n", mmap);
#endif
#endif
}

void
lm_init_framebuffer(void)
{
#ifndef __CONSOLE__
	s_pb_raspi_framebuffer();
#endif
}

void
lm_lock(void)
{
	/*  Nothing to do  */
}

void
lm_unlock(void)
{
	/*  Nothing to do  */
}

/*  Request redraw of a part of the screen  */
void
lm_redraw(int16_t x1, int16_t y1, int16_t width, int16_t height)
{
	int x, y, x2, y2, mx1, mx2, my1, my2, cx1, cx2, cy1, cy2, width2;
	pixel_t tpix;
	u_int32_t ofs, ofs2;
	
	x2 = x1 + width;
	y2 = y1 + height;
	if (my_graphic_mode == 1) {
		mx1 = my_origin_x;
		mx2 = mx1 + my_width / 2;
		my1 = my_origin_y;
		my2 = my1 + my_height / 2;
	} else {
		mx1 = my1 = 0;
		mx2 = my_width;
		my2 = my_height;
	}
	if (x1 < mx1)
		x1 = mx1;
	if (x2 > mx2)
		x2 = mx2;
	if (y1 < my1)
		y1 = my1;
	if (y2 > my2)
		y2 = my2;
	if (my_show_cursor) {
		cx1 = my_cursor_x * 8;
		cx2 = cx1 + 8;
		cy1 = my_cursor_y * 16;
		cy2 = cy1 + 16;
	} else cx1 = cx2 = cy1 = cy2 = -1;
	if (my_graphic_mode == 1) {
		ofs = y1 * my_width + x1;
		ofs2 = (y1 * 2 + my_origin_y) * my_width + (x1 * 2 + my_origin_x);
		width = x2 - x1;
		width2 = (x2 - x1) * 2;
	} else {
		ofs = ofs2 = y1 * my_width + x1;
		width = width2 = x2 - x1;
	}
#if __BAREMETAL__
	s_framebuffer = (char *)fb_addr;  /*  May change because of double buffering  */
#endif
	if (my_graphic_mode == 1) {
		/*  (2x2) expansion  */
		for (y = y1; y < y2; y++) {
			for (x = x1; x < x2; x++) {
				tpix = my_textscreen[ofs];
				if (tpix == 0 && x < cx2 && x >= cx1 && y < cy2 && y >= cy1)
					tpix = RGBFLOAT(0.5, 0.5, 0.5);  /*  cursor color: 50% gray  */
				if (tpix == 0)
					tpix = my_graphscreen[ofs];
				((pixel_t *)s_framebuffer)[ofs2] = tpix;
				((pixel_t *)s_framebuffer)[ofs2 + 1] = tpix;
				((pixel_t *)s_framebuffer)[ofs2 + my_width] = tpix;
				((pixel_t *)s_framebuffer)[ofs2 + my_width + 1] = tpix;
				ofs++;
				ofs2 += 2;
			}
			ofs += my_width - width;
			ofs2 += my_width * 2 - width2;
		}			
	} else {
		for (y = y1; y < y2; y++) {
			for (x = x1; x < x2; x++) {
				tpix = my_textscreen[ofs];
				if (tpix == 0 && x < cx2 && x >= cx1 && y < cy2 && y >= cy1)
					tpix = RGBFLOAT(0.5, 0.5, 0.5);  /*  cursor color: 50% gray  */
				if (tpix == 0)
					tpix = my_graphscreen[ofs];
				((pixel_t *)s_framebuffer)[ofs] = tpix;
				ofs++;
			}
			ofs += my_width - width;
		}
	}
}

#if 0
#pragma mark ====== Key Input ======
#endif

#ifndef __BAREMETAL__

/*  Key input.  */
/*  The standard input should be set to the RAW mode  */
int
lm_getch_raw(void)
{
	return tty_getch();
}

static struct termios s_SaveTermIos;
static struct termios s_RawTermIos;

int
tty_puts(const char *s)
{
	return fputs(s, stdout);
}

int
tty_putc(int c)
{
	return fputc(c, stdout);
}

int
tty_getch(void)
{
	u_int8_t c;
	int n;
	n = read(STDIN_FILENO, &c, 1);
	if (n > 0)
		return c;
	else return -1;	
}

void
set_raw_mode(void)
{
	tcgetattr(STDIN_FILENO, &s_SaveTermIos);
	s_RawTermIos = s_SaveTermIos;
	cfmakeraw(&s_RawTermIos);
	s_RawTermIos.c_cc[VMIN] = 1;
	s_RawTermIos.c_cc[VTIME] = 0;
	tcsetattr(STDIN_FILENO, 0, &s_RawTermIos);
	fcntl(STDIN_FILENO, F_SETFL, O_NONBLOCK);
	setvbuf(stdout, NULL, _IONBF, 0);
	tty_puts("\x1b[H");     /*  Set cursor home  */
	tty_puts("\x1b[J");     /*  Erase to end of screen  */
	tty_puts("\x1b[?25l");  /*  Hide cursor  */
}

void
unset_raw_mode(void)
{
	char buf[20];
	tcsetattr(STDIN_FILENO, 0, &s_SaveTermIos);
	snprintf(buf, sizeof buf, "\x1b[%d;1H", my_height / 16);  /*  Cursor down full screen  */
	tty_puts(buf);
	tty_puts("\x1b[?25h");  /*  Show cursor  */
}

#endif

#if 0
#pragma mark ====== main ======
#endif

int
main(int argc, const char **argv)
{
	extern int lua_main(void);

#ifndef __CONSOLE__
	if (argc > 1 && strcmp(argv[1], "-t") == 0)
		my_console = LM_CONSOLE_TTY;
#else
	my_console = LM_CONSOLE_TTY;
#endif
	
#ifdef __BAREMETAL__
	{
		extern int baremetal_init(void);
		baremetal_init();
	}
#else
	set_raw_mode();
	atexit(unset_raw_mode);
#endif

	lua_main();
	return 0;
}

void
bm_debug(int n)
{
	{ /* DEBUG */
		extern uint32_t fb_width, fb_height, fb_addr;
		u_int16_t *p, col;
		int i;
		extern void pb_init_framebuffer(void);
		extern void fb_init(int, int);
		fb_init(0, 0);
		my_width = fb_width;
		my_height = fb_height;
		p = (u_int16_t *)fb_addr;
		if (n == 0)
			col = 0xffff;
		else if (n == 1)
			col = 0xff00;
		else col = 0x00ff;
		for (i = 0; i < 200; i++) {
			*p = col;
			p += my_width + 1;
		}
	}
}